
#include <devices/gameport.h>
#include <clib/exec_protos.h>

#include "Joystick.h"

typedef struct GamePortTrigger GamePortTrigger;

STATIC VOID clearController(StdIO *io)
{
    UBYTE c = GPCT_NOCONTROLLER;

    io->io_Command = GPD_SETCTYPE;
    io->io_Data = (APTR)&c;
    io->io_Length = sizeof(UBYTE);
    io->io_Flags = IOF_QUICK;
    DoIO((IO *)io);
}

VOID closeJoystick(StdIO *io)
{
    MsgPort *mp = io->io_Message.mn_ReplyPort;

    if (CheckIO((IO *)io) == NULL)
    {
        AbortIO((IO *)io);
    }
    WaitIO((IO *)io);
    clearController(io);
    CloseDevice((IO *)io);
    DeleteIORequest((IO *)io);
    DeleteMsgPort(mp);
}

STATIC BOOL checkController(StdIO *io)
{
    UBYTE c;
    BOOL ready = FALSE;

    Forbid();

    io->io_Command = GPD_ASKCTYPE;
    io->io_Data = (APTR)&c;
    io->io_Length = sizeof(UBYTE);
    io->io_Flags = IOF_QUICK;
    DoIO((IO *)io);

    if (c == GPCT_NOCONTROLLER)
    {
        c = GPCT_ABSJOYSTICK;
        ready = TRUE;

        io->io_Command = GPD_SETCTYPE;
        io->io_Data = (APTR)&c;
        io->io_Length = sizeof(UBYTE);
        io->io_Flags = IOF_QUICK;
        DoIO((IO *)io);
    }

    Permit();
    return ready;
}

STATIC VOID setTrigger(StdIO *io, UWORD timeOut)
{
    GamePortTrigger gpt;

    gpt.gpt_Keys = GPTF_UPKEYS | GPTF_DOWNKEYS;
    gpt.gpt_XDelta = 1;
    gpt.gpt_YDelta = 1;
    gpt.gpt_Timeout = timeOut;

    io->io_Command = GPD_SETTRIGGER;
    io->io_Data = (APTR)&gpt;
    io->io_Length = sizeof(GamePortTrigger);
    io->io_Flags = IOF_QUICK;
    DoIO((IO *)io);
}

VOID clearIO(StdIO *io)
{
    io->io_Command = CMD_CLEAR;
    io->io_Data = (APTR)NULL;
    io->io_Length = 0;
    io->io_Flags = IOF_QUICK;
    DoIO((IO *)io);
}

VOID readEvent(StdIO *io, InputEvent *ie)
{
    io->io_Command = GPD_READEVENT;
    io->io_Data = (APTR)ie;
    io->io_Length = sizeof(InputEvent);
    io->io_Flags = 0;
    SendIO((IO *)io);
}

StdIO *openJoystick(UWORD timeOut, InputEvent *ie)
{
    MsgPort *mp;

    if (mp = CreateMsgPort())
    {
        StdIO *io;

        if (io = (StdIO *)CreateIORequest(mp, sizeof(StdIO)))
        {
            if (OpenDevice("gameport.device", 1, (IO *)io, 0) == 0)
            {
                if (checkController(io))
                {
                    setTrigger(io, timeOut);
                    clearIO(io);
                    readEvent(io, ie);                                        
                    return io;
                }
                CloseDevice((IO *)io);
            }
            DeleteIORequest((IO *)io);
        }
        DeleteMsgPort(mp);
    }
    return (StdIO *)NULL;
}
