
#include "RastPort.h"

#include <assert.h>

#include <graphics/rastport.h>
#include <graphics/clip.h>
#include <hardware/blit.h>
#include <hardware/custom.h>
#include <clib/graphics_protos.h>

#ifndef AMIGA
#define __far
#endif

__far extern struct Custom custom;

extern void c2p(register ULONG *buffer, register ULONG mask1, register ULONG mask2, register ULONG mask3, register ULONG mask4);

void drawIconQuick(RastPort *rp, Rectangle *rect, WORD ox, WORD oy, struct bltData *data)
{
    struct Custom *ptr = &custom;
    BitMap *src = data->src, *dest = rp->BitMap;
    UBYTE plane, depth = dest->Depth, mask = rp->Mask;
    WORD width = (rect->MaxX - rect->MinX + 16) >> 4;
    WORD height = rect->MaxY - rect->MinY + 1;
    WORD srcBpr = src->BytesPerRow;
    WORD destBpr = dest->BytesPerRow;
    LONG srcOffset = srcBpr * (data->sy + oy) + (((data->sx + ox) >> 4) << 1);
    LONG destOffset = destBpr * rect->MinY + ((rect->MinX >> 4) << 1);
    WORD srcMod = srcBpr - (width << 1);
    WORD destMod = destBpr - (width << 1);

    if (width == 0)
    {
        return;
    }

    for (plane = 0; plane < depth && mask; plane++)
    {
        if (mask & 1)
        {
            WaitBlit();
            ptr->bltcon0 = A_TO_D | SRCA | DEST;
            ptr->bltcon1 = 0;
            ptr->bltapt = src->Planes[plane] + srcOffset;
            ptr->bltdpt = dest->Planes[plane] + destOffset;
            ptr->bltamod = srcMod;
            ptr->bltdmod = destMod;
            ptr->bltafwm = 0xffff;
            ptr->bltalwm = 0xffff;
            ptr->bltsizv = height;
            ptr->bltsizh = width;
        }
        mask >>= 1;
    }
}

void drawDualIcon(RastPort *rp, Rectangle *rect, WORD ox, WORD oy, struct bltData *data)
{
    struct Custom *ptr = &custom;
    BitMap *src = data->src, *back = data->back, *dest = rp->BitMap;
    UBYTE plane, depth = dest->Depth, mask = rp->Mask, maskPlane = src->Depth - 1;
    WORD wordWidth = (rect->MaxX - rect->MinX + 16) >> 4;
    WORD height = rect->MaxY - rect->MinY + 1;
    WORD srcBpr = src->BytesPerRow;
    WORD backBpr = back->BytesPerRow;
    WORD destBpr = dest->BytesPerRow;
    LONG srcOffset = srcBpr * (data->sy + oy) + (((data->sx + ox) >> 4) << 1);	
    LONG backOffset = backBpr * data->by + ((data->bx >> 4) << 1);
    LONG destOffset = destBpr * rect->MinY + ((rect->MinX >> 4) << 1);
    WORD srcMod = srcBpr - (wordWidth << 1);
    WORD backMod = backBpr - (wordWidth << 1);
    WORD destMod = destBpr - (wordWidth << 1);
    BYTE shift = data->shift;

    if (wordWidth == 0)
    {
        return;
    }

    for (plane = 0; plane < depth && mask; plane++)
    {
        if (mask & 1)
        {
            WaitBlit();
            ptr->bltcon0 = 0xca | SRCA | SRCB | SRCC | DEST | (shift << ASHIFTSHIFT);
            ptr->bltcon1 = (shift << BSHIFTSHIFT);
            ptr->bltapt = src->Planes[maskPlane] + srcOffset;
            ptr->bltbpt = src->Planes[plane] + srcOffset;
            ptr->bltcpt = back->Planes[plane] + backOffset;
            ptr->bltdpt = dest->Planes[plane] + destOffset;
            ptr->bltamod = srcMod;
            ptr->bltbmod = srcMod;
            ptr->bltcmod = backMod;
            ptr->bltdmod = destMod;
            ptr->bltafwm = 0xffff;
            ptr->bltalwm = 0xffff << shift;
            ptr->bltsizv = height;
            ptr->bltsizh = wordWidth;
        }
        mask >>= 1;
    }
}

void drawDualIconLeft(RastPort *rp, Rectangle *rect, WORD ox, WORD oy, struct bltData *data)
{
    struct Custom *ptr = &custom;
    BitMap *src = data->src, *back = data->back, *dest = rp->BitMap;
    UBYTE plane, depth = dest->Depth, mask = rp->Mask, maskPlane = src->Depth - 1;
    WORD width = rect->MaxX - rect->MinX + 1;
    WORD wordWidth = (width + 15) >> 4;
    WORD height = rect->MaxY - rect->MinY + 1;
    WORD srcBpr = src->BytesPerRow;
    WORD backBpr = back->BytesPerRow;
    WORD destBpr = dest->BytesPerRow;
    LONG srcOffset = srcBpr * (data->sy + oy + height - 1) + (((data->sx + ox + width - 1) >> 4) << 1);
    LONG backOffset = backBpr * (data->by + height - 1) + (((data->bx + width - 1) >> 4) << 1);
    LONG destOffset = destBpr * (rect->MinY + height - 1) + (((rect->MinX + width - 1) >> 4) << 1);
    WORD srcMod = srcBpr - (wordWidth << 1);
    WORD backMod = backBpr - (wordWidth << 1);
    WORD destMod = destBpr - (wordWidth << 1);
    BYTE shift = data->shift;

    if (wordWidth == 0)
    {
        return;
    }

    for (plane = 0; plane < depth && mask; plane++)
    {
        if (mask & 1)
        {
            WaitBlit();
            ptr->bltcon0 = 0xca | SRCA | SRCB | SRCC | DEST | (shift << ASHIFTSHIFT);
            ptr->bltcon1 = BLITREVERSE | (shift << BSHIFTSHIFT);
            ptr->bltapt = src->Planes[maskPlane] + srcOffset;
            ptr->bltbpt = src->Planes[plane] + srcOffset;
            ptr->bltcpt = back->Planes[plane] + backOffset;
            ptr->bltdpt = dest->Planes[plane] + destOffset;
            ptr->bltamod = srcMod;
            ptr->bltbmod = srcMod;
            ptr->bltcmod = backMod;
            ptr->bltdmod = destMod;
            ptr->bltafwm = 0xffff;
            ptr->bltalwm = 0xffff >> shift;
            ptr->bltsizv = height;
            ptr->bltsizh = width;
        }
        mask >>= 1;
    }
}

void writeChunkyPixels(RastPort *rp, Rectangle *rect, WORD ox, WORD oy, struct wcpData *wcp)
{
    register ULONG mask1 = 0x0f0f0f0f, mask2 = 0xff00ff00, mask3 = 0x55555555, mask4 = 0x33333333;
    BitMap *bm = rp->BitMap;
    ULONG *buffer;
    WORD x, y;
    ULONG *planes[8];
    WORD p;
    WORD mod = (wcp->bpr - (rect->MaxX - rect->MinX + 1)) >> 2;

    for (p = 0; p < bm->Depth; p++)
    {
        planes[p] = (ULONG *)bm->Planes[p];
    }

    buffer = (ULONG *)(wcp->buffer + (wcp->bpr * oy) + ox);

    for (y = rect->MinY; y <= rect->MaxY; y++)
    {
        for (x = rect->MinX; x <= rect->MaxX; x += 32)
        {
            c2p(buffer, mask1, mask2, mask3, mask4);

            for (p = 0; p < bm->Depth; p++)
            {
                *planes[p]++ = buffer[p << 1];
            }
            for (p = 4; p < bm->Depth; p++)
            {
                *planes[p]++ = buffer[((p - 4) << 1) + 1];
            }
            buffer += 8;
        }
        buffer += mod;
    }
}

void clipBlit(RastPort *rp, Rectangle *opRect, void (*blitOp)(RastPort *rp, Rectangle *rect, WORD ox, WORD oy, APTR data), APTR data)
{
    Layer *l = rp->Layer;
    ClipRect *cr;
    Rectangle rect, clip, *bounds;

    OwnBlitter();

    if (l == NULL)
    {
        (*blitOp)(rp, opRect, 0, 0, data);
    }
    else
    {
        bounds = &l->bounds;
        rect.MinX = opRect->MinX + bounds->MinX;
        rect.MinY = opRect->MinY + bounds->MinY;
        rect.MaxX = opRect->MaxX + bounds->MinX;
        rect.MaxY = opRect->MaxY + bounds->MinY;

        for (cr = l->ClipRect; cr != NULL; cr = cr->Next)
        {
            if (cr->lobs == NULL)
            {
                clip.MinX = Max(rect.MinX, cr->bounds.MinX);
                clip.MinY = Max(rect.MinY, cr->bounds.MinY);
                clip.MaxX = Min(rect.MaxX, cr->bounds.MaxX);
                clip.MaxY = Min(rect.MaxY, cr->bounds.MaxY);

                if (clip.MinX <= clip.MaxX && clip.MinY <= clip.MaxY)
                {
                    (*blitOp)(rp, &clip, clip.MinX - rect.MinX, clip.MinY - rect.MinY, data);
                }
            }
        }
    }

    DisownBlitter();
}
